# -*- coding: iso-8859-1; mode: python -*-
# -----------------------------------------------------------
# This file is part of DRMAA-python.
# Copyright (C) 2005 StatPro Italia s.r.l. Milano 
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.
# -----------------------------------------------------------
#  Author: Enrico Sirola <enrico.sirola@statpro.com> 
# 
#  $Id$
# -----------------------------------------------------------

"""A Python DRMAA interface"""

import cDRMAA
import exceptions
from inspect import isfunction, isbuiltin
from re import compile
from datetime import datetime, time

_DATETIME_FMT = "%Y/%m/%d %H:%M %z"
_TIME_FMT = "%H:%M %z"

def _fromVDict(vdict):
    rv = {}
    for v in vdict:
        key, val = v.split('=')
        rv[key] = val
    return rv

def _toVDict(map):
    vdict = []
    for k in map:
        vdict.append("%s=%s" % (k, map[k]))
    return vdict

def _get_values_list(struct, get_next, deallocate):
    try:
        values_list = []
        while 1:
            no_more_values, value = get_next(struct)
            if no_more_values: break
            else:
                values_list.append(value)
        return values_list
    finally:
        deallocate(struct)

def parseTime(t_string):
    match=parseTime.DATETIME_RE.match(t_string)
    if match:
        year,month,day,hour,minute=[int(x) for x in match.groups()]
        if year < parseTime.YEAR_OFFSET: year+=2000
        return datetime(year=year,month=month,day=day,hour=hour,minute=minute)
    match=parseTime.TIME_RE.match(t_string)
    if match:
        hour,minute=[int(x) for x in match.groups()]
        return time(hour=hour,minute=minute)
    else:
        raise "oops!"
parseTime.TIME_RE=compile("^(\d{2}):(\d{2})")
parseTime.DATETIME_RE=compile("^(\d{2,4})/(\d{2})/(\d{2}) (\d{2}):(\d{2})")
parseTime.YEAR_OFFSET=70


class DrmaaError(exceptions.Exception): pass
class InitError(DrmaaError): pass
class ExitError(DrmaaError): pass
class AuthorizationError(DrmaaError): pass
class InvalidContactStringError(DrmaaError): pass
class DefaultContactStringError(DrmaaError): pass
class DeniedByDrmError(DrmaaError): pass
class DrmCommunicationError(DrmaaError): pass
class DrmsExitError(DrmaaError): pass
class InconsistentStateError(DrmaaError): pass
class HoldInconsistentStateError(InconsistentStateError): pass
class ReleaseInconsistentStateError(InconsistentStateError): pass
class ResumeInconsistentStateError(InconsistentStateError): pass
class SuspendInconsistentStateError(InconsistentStateError): pass
class DrmsInitError(DrmaaError): pass
class InvalidJobError(DrmaaError): pass
class InvalidAttributeError(DrmaaError): pass
class ConflictingAttributeValuesError(InvalidAttributeError): pass
class InvalidAttributeFormatError(InvalidAttributeError): pass
class InvalidAttributeValueError(InvalidAttributeError): pass
class NoResourceUsageError(DrmaaError): pass
class ExitTimeoutError(DrmaaError): pass
class SessionError(DrmaaError): pass
class NoActiveSessionError(SessionError): pass
class AlreadyActiveSessionError(SessionError): pass
class RuntimeError(exceptions.RuntimeError): pass
class InternalError(RuntimeError): pass
class UnsupportedAttributeError(RuntimeError): pass
class InvalidJobTemplateError(RuntimeError): pass
class NoMemoryError(RuntimeError): pass
class InvalidArgumentError(DrmaaError): pass
class NoRusageError(RuntimeError): pass
class TryLaterError(RuntimeError): pass

_cerrors=(cDRMAA.DRMAA_ERRNO_INTERNAL_ERROR,
          cDRMAA.DRMAA_ERRNO_DRM_COMMUNICATION_FAILURE,
          cDRMAA.DRMAA_ERRNO_AUTH_FAILURE,
          cDRMAA.DRMAA_ERRNO_INVALID_ARGUMENT,
          cDRMAA.DRMAA_ERRNO_NO_ACTIVE_SESSION,
          cDRMAA.DRMAA_ERRNO_NO_MEMORY,
          cDRMAA.DRMAA_ERRNO_INVALID_CONTACT_STRING,
          cDRMAA.DRMAA_ERRNO_DEFAULT_CONTACT_STRING_ERROR,
          cDRMAA.DRMAA_ERRNO_DRMS_INIT_FAILED,
          cDRMAA.DRMAA_ERRNO_ALREADY_ACTIVE_SESSION,
          cDRMAA.DRMAA_ERRNO_DRMS_EXIT_ERROR,
          cDRMAA.DRMAA_ERRNO_INVALID_ATTRIBUTE_FORMAT,
          cDRMAA.DRMAA_ERRNO_INVALID_ATTRIBUTE_VALUE,
          cDRMAA.DRMAA_ERRNO_CONFLICTING_ATTRIBUTE_VALUES,
          cDRMAA.DRMAA_ERRNO_TRY_LATER,
          cDRMAA.DRMAA_ERRNO_DENIED_BY_DRM,
          cDRMAA.DRMAA_ERRNO_INVALID_JOB,
          cDRMAA.DRMAA_ERRNO_RESUME_INCONSISTENT_STATE,
          cDRMAA.DRMAA_ERRNO_SUSPEND_INCONSISTENT_STATE,
          cDRMAA.DRMAA_ERRNO_HOLD_INCONSISTENT_STATE,
          cDRMAA.DRMAA_ERRNO_RELEASE_INCONSISTENT_STATE,
          cDRMAA.DRMAA_ERRNO_EXIT_TIMEOUT,
          cDRMAA.DRMAA_ERRNO_NO_RUSAGE,)

_errMin=min(_cerrors)
_errMax=max(_cerrors)

_errindex = [0]*(_errMax-_errMin+1)
_errindex[cDRMAA.DRMAA_ERRNO_INTERNAL_ERROR-_errMin]=InternalError
_errindex[cDRMAA.DRMAA_ERRNO_DRM_COMMUNICATION_FAILURE-_errMin]=DrmCommunicationError
_errindex[cDRMAA.DRMAA_ERRNO_AUTH_FAILURE-_errMin]=AuthorizationError
_errindex[cDRMAA.DRMAA_ERRNO_INVALID_ARGUMENT-_errMin]=InvalidArgumentError
_errindex[cDRMAA.DRMAA_ERRNO_NO_ACTIVE_SESSION-_errMin]=NoActiveSessionError
_errindex[cDRMAA.DRMAA_ERRNO_NO_MEMORY-_errMin]=NoMemoryError
_errindex[cDRMAA.DRMAA_ERRNO_INVALID_CONTACT_STRING-_errMin]=InvalidContactStringError
_errindex[cDRMAA.DRMAA_ERRNO_DEFAULT_CONTACT_STRING_ERROR-_errMin]=DefaultContactStringError
_errindex[cDRMAA.DRMAA_ERRNO_DRMS_INIT_FAILED-_errMin]=InitError
_errindex[cDRMAA.DRMAA_ERRNO_ALREADY_ACTIVE_SESSION-_errMin]=AlreadyActiveSessionError
_errindex[cDRMAA.DRMAA_ERRNO_DRMS_EXIT_ERROR-_errMin]=DrmsExitError
_errindex[cDRMAA.DRMAA_ERRNO_INVALID_ATTRIBUTE_FORMAT-_errMin]=InvalidAttributeFormatError
_errindex[cDRMAA.DRMAA_ERRNO_INVALID_ATTRIBUTE_VALUE-_errMin]=InvalidAttributeValueError
_errindex[cDRMAA.DRMAA_ERRNO_CONFLICTING_ATTRIBUTE_VALUES-_errMin]=ConflictingAttributeValuesError
_errindex[cDRMAA.DRMAA_ERRNO_TRY_LATER-_errMin]=TryLaterError
_errindex[cDRMAA.DRMAA_ERRNO_DENIED_BY_DRM-_errMin]=DeniedByDrmError
_errindex[cDRMAA.DRMAA_ERRNO_INVALID_JOB-_errMin]=InvalidJobError
_errindex[cDRMAA.DRMAA_ERRNO_RESUME_INCONSISTENT_STATE-_errMin]=ResumeInconsistentStateError
_errindex[cDRMAA.DRMAA_ERRNO_SUSPEND_INCONSISTENT_STATE-_errMin]=SuspendInconsistentStateError
_errindex[cDRMAA.DRMAA_ERRNO_HOLD_INCONSISTENT_STATE-_errMin]=HoldInconsistentStateError
_errindex[cDRMAA.DRMAA_ERRNO_RELEASE_INCONSISTENT_STATE-_errMin]=ReleaseInconsistentStateError
_errindex[cDRMAA.DRMAA_ERRNO_EXIT_TIMEOUT-_errMin]=ExitTimeoutError
_errindex[cDRMAA.DRMAA_ERRNO_NO_RUSAGE-_errMin]=NoRusageError

def errParse(errno, errstring=''):
    if errno==cDRMAA.DRMAA_ERRNO_SUCCESS: return
    else: raise _errindex[errno-_errMin](errstring)

class Session(object):
    
    def init(self, contactString=''):
        """The init() method is used to initialize a DRMAA session for
use. The contact parameter is an implementation-dependent string that
may be used to specify which DRM system to use. This method must be
called before any other DRMAA calls, except for getDrmSystem(),
getDrmaaImplementation(), and getContact().  If contact is null, the
default DRM system is used, provided there is only one DRMAA
implementation available. If there is more than one DRMAA
implementation available, init() throws a
NoDefaultContactStringSelectedException. init() should be called only
once, by only one of the threads. The main thread is recommended. A
call to init() by another thread or additional calls to init() by the
same thread with throw a SessionAlreadyActiveException.  """
        eno, estr = cDRMAA.drmaa_init(contactString)
        self._is_open=True
        errParse(eno, estr)
    def exit(self):
        """The exit() is used to disengage from DRM and allow the
DRMAA implementation to perform any necessary internal cleanup. This
routine ends the current DRMAA session but doesn't affect any jobs
(e.g., queued and running jobs remain queued and running). exit()
should be called only once, by only one of the threads. Additional
calls to exit() beyond the first will throw a
NoActiveSessionException.
"""
        eno, estr = cDRMAA.drmaa_exit()
        self._is_open=False
        errParse(eno, estr)

    def createJobTemplate(self):
        """The createJobTemplate() method is used to get a new job
template. The job template is used to set the environment for jobs to
be submitted. Once the job template has been created, it should also
be deleted (via deleteJobTemplate()) when no longer needed. Failure to
do so may result in a memory leak.  """
        eno, jt, estr = cDRMAA.drmaa_allocate_job_template()
        errParse(eno, estr)
        return JobTemplate(jt)
    def deleteJobTemplate(self, jobTemplate):
        """The deleteJobTemplate() method is used to deallocate a job
template. This routine has no effect on running jobs.  """
        eno, estr = cDRMAA.drmaa_delete_job_template(jobTemplate._jt)
        errParse(eno, estr)

    def runJob(self, jobTemplate):
        """The runJob() method is used to submit a job with attributes
defined in the job template, jt. The returned job identifier is a
String identical to that returned from the underlying DRM system.  """
        eno, jobId, estr = cDRMAA.drmaa_run_job(jobTemplate._jt)
        errParse(eno, estr)
        return jobId
    def runBulkJobs(self, jobTemplate, beginIndex, endIndex, step):
        """The runBulkJobs() method is used to submit a set of
parametric jobs, dependent on the implied loop index, each with
attributes defined in the job template, jt. The returned job
identifiers are Strings identical to those returned from the
underlying DRM system.  The JobTemplate class defines a
PARAMETRIC_INDEX placeholder for use in specifying paths.  This
placeholder is used to represent the individual identifiers of the
tasks submitted through this method.  """
        eno, jobIds, estr = cDRMAA.drmaa_run_bulk_jobs(jobTemplate._jt, beginIndex, endIndex, step)
        errParse(eno, estr)
        try:
            ids = []
            idx = beginIndex
            while idx <= endIndex:
                eno, jid = cDRMAA.drmaa_get_next_job_id(jobIds)
                errParse(eno)
                ids.append(jid)
                idx+=step
            return ids
        finally:
            cDRMAA.drmaa_release_job_ids(jobIds)
    # job control
    SUSPEND = cDRMAA.DRMAA_CONTROL_SUSPEND
    RESUME = cDRMAA.DRMAA_CONTROL_RESUME
    HOLD = cDRMAA.DRMAA_CONTROL_HOLD
    RELEASE = cDRMAA.DRMAA_CONTROL_RELEASE
    TERMINATE =cDRMAA.DRMAA_CONTROL_TERMINATE
    def control(self, jobName, operation):
        """The control() method is used to hold, release, suspend,
resume, or kill the job identified by jobId.  If jobId is
JOB_IDS_SESSION_ALL, then this routine acts on all jobs submitted
during this DRMAA session up to the moment control() is called. To
avoid thread races in multithreaded applications, the DRMAA
implementation user should explicitly synchronize this call with any
other job submission calls or control calls that may change the number
of remote jobs.  The legal values for action and their meanings are:
          * SUSPEND: stop the job,
          * RESUME: (re)start the job,
          * HOLD: put the job on-hold,
          * RELEASE: release the hold on the job, and
          * TERMINATE: kill the job.
This method returns once the action has been acknowledged by the DRM
system, but does not necessarily wait until the action has been
completed.  Some DRMAA implementations may allow this method to be
used to control jobs submitted external to the DRMAA session, such as
jobs submitted by other DRMAA session in other DRMAA implementations
or jobs submitted via native utilities.
"""
        eno, estr = cDRMAA.drmaa_control(
            jobName, operation)
        errParse(eno, estr)

    # timeout constants
    TIMEOUT_WAIT_FOREVER = cDRMAA.DRMAA_TIMEOUT_WAIT_FOREVER
    TIMEOUT_NO_WAIT = cDRMAA.DRMAA_TIMEOUT_NO_WAIT
    def synchronize(self, jobList, timeout, dispose):
        """This method waits until all jobs specified by jobIds have
finished execution. If jobIds contains JOB_IDS_SESSION_ALL, then this
method waits for all jobs submitted during this DRMAA session up to
the moment synchronize() is called. To avoid thread race conditions in
multithreaded applications, the DRMAA implementation user should
explicitly synchronize this call with any other job submission calls
or control calls that may change the number of remote jobs.  To
prevent blocking indefinitely in this call, the caller may use a
timeout specifying after how many seconds to block in this call. The
value TIMEOUT_WAIT_FOREVER may be specified to wait indefinitely for a
result. The value TIMEOUT_NO_WAIT may be specified to return
immediately if no result is available. If the call exits before the
timeout has elapsed, all the jobs have been waited on or there was an
interrupt. If the invocation exits on timeout, an ExitTimeoutException
is thrown. The caller should check system time before and after this
call in order to be sure of how much time has passed.  The dispose
parameter specifies how to treat the reaping of the remote job's
internal data record, which includes a record of the job's consumption
of system resources during its execution and other statistical
information. If set to true, the DRM will dispose of the job's data
record at the end of the synchroniize() call. If set to false, the
data record will be left for future access via the wait() method.
"""
        if dispose: d = 1
        else: d = 0
        eno, estr=cDRMAA.drmaa_synchronize(jobList, timeout, d)
        errParse(eno, estr)
    def wait(self, jobName, timeout):
        """This method will wait for a job with jobId to finish
execution or fail. If the special string, JOB_IDS_SESSION_ANY, is
provided as the jobId, this routine will wait for any job from the
session. This routine is modeled on the wait3 POSIX routine.  The
timeout value is used to specify the desired behavior when a result is
not immediately available. The value, TIMEOUT_WAIT_FOREVER, may be
specified to wait indefinitely for a result. The value,
TIMEOUT_NO_WAIT, may be specified to return immediately if no result
is available. Alternatively, a number of seconds may be specified to
indicate how long to wait for a result to become available.  If the
call exits before timeout, either the job has been waited on
successfully or there was an interrupt. If the invocation exits on
timeout, an ExitTimeoutException is thrown. The caller should check
system time before and after this call in order to be sure how much
time has passed.  The routine reaps job data records on a successful
call, so any subsequent calls to wait() will fail, throwing an
InvalidJobException, meaning that the job's data record has been
already reaped.  This exception is the same as if the job were
unknown. (The only case where wait() can be successfully called on a
single job more than once is when the previous call to wait() timed
out before the job finished.)  """
        eno, id, stat, rusage, estr = cDRMAA.drmaa_wait(jobName, timeout)
        errParse(eno, estr)
        try:
            eno, aborted, estr = cDRMAA.drmaa_wifaborted(stat)
            termsig=None
            if aborted: aborted=True
            else: aborted=False
            eno, exited, estr = cDRMAA.drmaa_wifexited(stat)
            if exited: exited=True
            else: exited=False
            if exited:
                eno, exitstatus, estr=cDRMAA.drmaa_wexitstatus(stat)
                signaled=False
            else:
                eno, signaled, estr=cDRMAA.drmaa_wifsignaled(stat)
                if signaled: signaled=True
                else: signaled=False
                if signaled:
                    eno, termsig, estr=cDRMAA.drmaa_wtermsig(stat)
                exitstatus=None
            usage_list=[]
        except:
            cDRMAA.drmaa_release_attr_values(rusage)
            raise
        return JobInfo(id,
                       exitstatus,
                       _fromVDict(_get_values_list(rusage,
                                                   cDRMAA.drmaa_get_next_attr_value,
                                                   cDRMAA.drmaa_release_attr_values)),
                       exited,
                       signaled,
                       termsig,
                       None,
                       aborted)
    # job status
    UNDETERMINED = cDRMAA.DRMAA_PS_UNDETERMINED
    QUEUED_ACTIVE = cDRMAA.DRMAA_PS_QUEUED_ACTIVE
    SYSTEM_ON_HOLD = cDRMAA.DRMAA_PS_SYSTEM_ON_HOLD
    USER_ON_HOLD = cDRMAA.DRMAA_PS_USER_ON_HOLD
    USER_SYSTEM_ON_HOLD = cDRMAA.DRMAA_PS_USER_SYSTEM_ON_HOLD
    RUNNING = cDRMAA.DRMAA_PS_RUNNING
    SYSTEM_SUSPENDED = cDRMAA.DRMAA_PS_SYSTEM_SUSPENDED
    USER_SUSPENDED = cDRMAA.DRMAA_PS_USER_SUSPENDED
    USER_SYSTEM_SUSPENDED = cDRMAA.DRMAA_PS_USER_SYSTEM_SUSPENDED
    DONE = cDRMAA.DRMAA_PS_DONE
    FAILED = cDRMAA.DRMAA_PS_FAILED
    def getJobProgramStatus(self, jobName):
        """The getJobProgramStatus() method returns the program status
of the job identified by jobId.  The possible values returned from
this method are:
       * UNDETERMINED: process status cannot be determined,
       * QUEUED_ACTIVE: job is queued and active,
       * SYSTEM_ON_HOLD: job is queued and in system hold,
       * USER_ON_HOLD: job is queued and in user hold,
       * USER_SYSTEM_ON_HOLD: job is queued and in user and system hold,
       * RUNNING: job is running,
       * SYSTEM_SUSPENDED: job is system suspended,
       * USER_SUSPENDED: job is user suspended,
       * DONE: job finished normally, and
       * FAILED: job finished, but failed.
The DRMAA implementation should always get the status of the job from
the DRM system unless the status has already been determined to be
FAILED or DONE and the status has been successfully cached. Terminated
jobs return a FAILED status.
"""
        eno, ps, estr = cDRMAA.drmaa_job_ps(jobName)
        errParse(eno, estr)
        return ps
    def getContact(self):
        """If called before init(), this method returns a comma
delimited String containing the contact Strings available from the
default DRMAA implementation, one element per DRM system available. If
called after init(), this method returns the contact String for the
DRM system to which the session is attached. The returned String is
implementation dependent.
"""
        eno, cnt, estr = cDRMAA.drmaa_get_contact()
        errParse(eno, estr)
        return cnt
    contact=property(getContact)
    def getVersion(self):
        """This method returns a Version object containing the major
and minor version numbers of the DRMAA library. For DRMAA 1.0, major
is 1 and minor is 0.  """
        eno, major, minor, estr = cDRMAA.drmaa_version()
        errParse(eno, estr)
        return major, minor
    version=property(getVersion)
    def getDRMSInfo(self):
        """If called before init(), this method returns a comma
delimited list of DRM systems, one element per DRM system
implementation provided. If called after init(), this method returns
the selected DRM system. The returned String is implementation
dependent.  """
        eno, system, estr = cDRMAA.drmaa_get_DRMAA_implementation()
        errParse(eno, estr)
        return system
    DRMSInfo=property(getDRMSInfo)
    def getDRMAAImplementation(self):
        """If called before init(), this method returns a comma
delimited list of DRMAA implementations, one element for each DRMAA
implementation provided. If called after init(), this method returns
the selected DRMAA implementation. The returned String is
implementation dependent and may contain the DRM system as a
component.  """
        eno, impl, estr = cDRMAA.drmaa_get_DRMAA_implementation()
        errParse(eno, estr)
        return impl
    DRMAAImplementation=property(getDRMAAImplementation)
    def __del__(self):
        if getattr(self, '_is_open', False):
            self.exit()


class JobTemplate(object):
    """In order to define the attributes associated with a job, a
DRMAA application uses the JobTemplate class. JobTemplates are created
via the active Session implementation. A DRMAA application gets a
JobTemplate from the active Session, specifies in the JobTemplate any
required job parameters, and the passes the JobTemplate back to the
Session when requesting that a job be executed. When finished, the
DRMAA application should call the Session.deleteJobTemplate() method
to allow the underlying implementation to free any resources bound to
the JobTemplate object.  """
    HOME_DIRECTORY = cDRMAA.DRMAA_PLACEHOLDER_HD
    PARAMETRIC_INDEX = cDRMAA.DRMAA_PLACEHOLDER_INCR
    WORKING_DIRECTORY = cDRMAA.DRMAA_PLACEHOLDER_WD
    ACTIVE = cDRMAA.DRMAA_SUBMISSION_STATE_ACTIVE
    HOLD = cDRMAA.DRMAA_SUBMISSION_STATE_HOLD
    
    def __getter(self, id):
        eno, value, estr = cDRMAA.drmaa_get_attribute(
            self._jt, id)
        errParse(eno, estr)
        return value
    def __setter(self, id, value):
        eno, estr = cDRMAA.drmaa_set_attribute(
            self._jt, id, value)
        errParse(eno, estr)

    def __vgetter(self, id):
        eno, v, estr = cDRMAA.drmaa_get_vector_attribute(
            self._jt, id)
        try:
            errParse(eno, estr)
        except InvalidAttributeValueError:
            v = None
        return v
    def __vsetter(self, id, value):
        eno, estr = cDRMAA.drmaa_set_vector_attribute(
            self._jt, id, value)
        errParse(eno, estr)

    def __timesetter(self, id, time):
        if hasattr(time, "year"):
            self.__setter(id, time.strftime(_DATETIME_FMT))
        elif hasattr(time, "hour"):
            self.__setter(id, time.strftime(_TIME_FMT))
        else:
            self.__setter(id, time)
    def __timegetter(self, id):
        ctime = self.__getter(id)
        return parseTime(ctime)
        
    def __boolsetter(self, id, value):
        if value: cvalue='y'
        else: cvalue='n'
        self.__setter(id, cvalue)
    def __boolgetter(self, id):
        cvalue = self.__getter(id)
        if cvalue == 'y': return True
        else: return False
        
    def __init__(self, jt):
        self._jt=jt

    def setRemoteCommand(self, command):
        self.__setter(cDRMAA.DRMAA_REMOTE_COMMAND, command)
    def getRemoteCommand(self):
        return self.__getter(cDRMAA.DRMAA_REMOTE_COMMAND)
    remoteCommand=property(getRemoteCommand, setRemoteCommand)

    def setArgs(self, args):
        self.__vsetter(cDRMAA.DRMAA_V_ARGV, args),             
    def getArgs(self):
        return _get_values_list(self.__vgetter(cDRMAA.DRMAA_V_ARGV),
                                cDRMAA.drmaa_get_next_attr_value,
                                cDRMAA.drmaa_release_attr_values)
    args=property(getArgs, setArgs)

    def setSubmissionState(self, state):
        self.__setter(cDRMAA.DRMAA_JS_STATE, state)
    def getSubmissionState(self):
        return self.__getter(cDRMAA.DRMAA_JS_STATE)
    submissionState=property(setSubmissionState, getSubmissionState)

    def setEnvironment(self, env):
        self.__vsetter(cDRMAA.DRMAA_V_ENV, _toVDict(env))
    def getEnvironment(self):
        return _fromVDict(_get_values_list(self.__vgetter(cDRMAA.DRMAA_V_ENV),
                                           cDRMAA.drmaa_get_next_attr_value,
                                           cDRMAA.drmaa_release_attr_values))
    environment=property(getEnvironment, setEnvironment)

    def setWorkingDirectory(self, directory):
        self.__setter(cDRMAA.DRMAA_WD, directory)
    def getWorkingDirectory(self):
        return self.__getter(cDRMAA.DRMAA_WD)
    workingDirectory=property(getWorkingDirectory, setWorkingDirectory)

    def getCategory(self):
        self.__getter(cDRMAA.DRMAA_JOB_CATEGORY)
    category=property(getCategory)

    def setNativeSpecification(self, spec):
        self.__setter(cDRMAA.DRMAA_NATIVE_SPECIFICATION, spec)
    def getNativeSpecification(self):
        return self.__getter(cDRMAA.DRMAA_NATIVE_SPECIFICATION)
    nativeSpecification=property(getNativeSpecification, setNativeSpecification)

    def setEmail(self, email):
        self.__vsetter(cDRMAA.DRMAA_V_EMAIL, email)
    def getEmail(self):
        return _get_values_list(self.__vgetter(cDRMAA.DRMAA_V_EMAIL),
                                cDRMAA.drmaa_get_next_attr_value,
                                cDRMAA.drmaa_release_attr_values)
    email=property(getEmail, setEmail)

    def setBlockEmail(self, blockEmail):
        self.__boolsetter(cDRMAA.DRMAA_BLOCK_EMAIL, blockEmail)
    def getBlockEmail(self):
        return self.__boolgetter(cDRMAA.DRMAA_BLOCK_EMAIL)
    blockEmail=property(getBlockEmail, setBlockEmail)

    def setStartTime(self, startTime):
        self.__timesetter(cDRMAA.DRMAA_START_TIME, startTime)
    def getStartTime(self):
        return self.__timegetter(cDRMAA.DRMAA_START_TIME)
    startTime=property(getStartTime, setStartTime)

    def setName(self, jname):
        self.__setter(cDRMAA.DRMAA_JOB_NAME, jname)
    def getName(self):
        return self.__getter(cDRMAA.DRMAA_JOB_NAME)
    name=property(getName, setName)

    def setInputPath(self, ipath):
        self.__setter(cDRMAA.DRMAA_INPUT_PATH, ipath)
    def getInputPath(self):
        return self.__getter(cDRMAA.DRMAA_INPUT_PATH)
    inputPath=property(getInputPath, setInputPath)

    def setErrorPath(self, errorPath):
        self.__setter(cDRMAA.DRMAA_ERROR_PATH, errorPath)
    def getErrorPath(self):
        return self.__getter(cDRMAA.DRMAA_ERROR_PATH)
    errorPath=property(getErrorPath, setErrorPath)

    def setOutputPath(self, outputPath):
        self.__setter(cDRMAA.DRMAA_OUTPUT_PATH, outputPath)
    def getOutputPath(self):
        return self.__getter(cDRMAA.DRMAA_OUTPUT_PATH)
    outputPath=property(getOutputPath, setOutputPath)

    def setJoinFiles(self, joinFiles):
        self.__boolsetter(cDRMAA.DRMAA_JOIN_FILES, joinFiles)
    def getJoinFiles(self):
        return self.__boolgetter(cDRMAA.DRMAA_JOIN_FILES)
    joinFiles=property(getJoinFiles, setJoinFiles)

    def setTransferFiles(self, transferMode):
        self.__setter(cDRMAA.DRMAA_TRANSFER_FILES, transferMode)
    def getTransferFiles(self):
        return self.__getter(cDRMAA.DRMAA_TRANSFER_FILES)
    transferFiles=property(getTransferFiles, setTransferFiles)

    def setDeadlineTime(self, deadline):
        self.__timesetter(cDRMAA.DRMAA_DEADLINE_TIME, deadline)
    def getDeadlineTime(self):
        return self.__timegetter(cDRMAA.DRMAA_DEADLINE_TIME)
    deadlineTime=property(getDeadlineTime, setDeadlineTime)

    def setHardWallclockTimeLimit(self, limit):
        self.__timesetter(cDRMAA.DRMAA_WCT_HLIMIT, limit)
    def getHardWallclockTimeLimit(self):
        return self.__getter(cDRMAA.DRMAA_WCT_HLIMIT)
    hardWallclockTimeLimit=property(getHardWallclockTimeLimit,
                                    setHardWallclockTimeLimit)

    def setSoftWallclockTimeLimit(self, limit):
        self.__timesetter(cDRMAA.DRMAA_WCT_SLIMIT, limit)
    def getSoftWallclockTimeLimit(self):
        return self.__getter(cDRMAA.DRMAA_WCT_SLIMIT)
    softWallclockTimeLimit=property(getSoftWallclockTimeLimit,
                                    setSoftWallclockTimeLimit)

    def setHardRunDurationLimit(self, limit):
        self.__timesetter(cDRMAA.DRMAA_DURATION_HLIMIT, limit)
    def getHardRunDurationLimit(self):
        return self.__getter(cDRMAA.DRMAA_DURATION_HLIMIT)
    hardRunDurationLimit=property(getHardRunDurationLimit,
                                  setHardRunDurationLimit)

    def setSoftRunDurationLimit(self, limit):
        self.__timesetter(cDRMAA.DRMAA_DURATION_SLIMIT, limit)
    def getSoftRunDurationLimit(self):
        return self.__getter(cDRMAA.DRMAA_DURATION_SLIMIT)
    softRunDurationLimit=property(getSoftRunDurationLimit,
                                  setSoftRunDurationLimit)

    def getAttributeNames(self):
        eno, names, estr = cDRMAA.drmaa_get_attribute_names()
        errParse(eno, estr)
        eno, vnames, estr = cDRMAA.drmaa_get_vector_attribute_names()
        errParse(eno, estr)        
        return _get_values_list(names,
                                cDRMAA.drmaa_get_next_attr_name,
                                cDRMAA.drmaa_release_attr_names) + \
               _get_values_list(vnames,
                                cDRMAA.drmaa_get_next_attr_name,
                                cDRMAA.drmaa_release_attr_names)
    attributeNames=property(getAttributeNames)

class JobInfo(object):
    def __init__(self,
                 name,
                 statusCode,
                 resourceUsage,
                 ifExited,
                 ifSignaled,
                 terminatingSignal,
                 coreDump,
                 ifAborted):
        self.name=name
        self.statusCode=statusCode
        self.resourceUsage=resourceUsage
        self.ifExited=ifExited
        self.ifSignaled=ifSignaled
        self.terminatingSignal=terminatingSignal
        self.coreDump=coreDump
        self.ifAborted=ifAborted
    def getJobId(self):
        """This method returns the completed job's id."""
        return self.name
    def getResourceUsage(self):
        """This method returns the completed job's resource usage data."""
        return self.resourceUsage
    def ifExited(self):
        """This method returns True if the job terminated
normally. False can also indicate that although the job has terminated
normally an exit status is not available or that it is not known
whether the job terminated normally. In both cases getExitStatus()
doesn't provide exit status information.  True indicates more detailed
diagnosis can be provided by means of getExitStatus()."""
        return self.ifExited
    def getExitStatus(self):
        """If ifExited() returns True, this function returns the exit
code that the job passed to _exit() (see exit(2)) or exit(3C)), or the
value that the child process returned from main."""
        return self.statusCode
    def ifSignaled(self):
        """The ifSignaled() method returns True if the job terminated
due to the receipt of a signal. False can also indicate that although
the job has terminated due to the receipt of a signal, the signal is
not available or that it is not known whether the job terminated due
to the receipt of a signal. In both cases getTerminatingSignal() does
not provide signal information."""
        return self.ifSignaled
    def getTerminatingSignal(self):
        """If ifSignaled() returns true, this method returns a
representation of the signal that caused the termination of the
job. For signals declared by POSIX, the symbolic names are be returned
(e.g., SIGABRT, SIGALRM).  For signals not declared by POSIX, a DRM
dependent string is returned.  """
        return self.terminatingSignal
    def coreDump(self):
        """If ifSignaled() returns True, this function returns True if
a core image of the terminated job was created."""
        return self.coreDump
    def ifAborted(self):
        """This method returns true if the job ended before entering
the running state."""
        return self.ifAborted

# class FileTransferMode(object):
#     def setInputStream(self, istream):
#         pass
#     def getInputStream(self):
#         pass
#     def setOutputStream(self, ostream):
#         pass
#     def getOutputStream(self):
#         pass
#     def setErrorStream(self, estream):
#         pass
#     def getErrorStream(self):
#         pass
#     def clone(self):
#         pass
#     def toString(self):
#         pass
#     def equals(self, o):
#         pass
#     def hashCode(self):
#         pass


__all__ = [ 'TryLaterError', 'NoRusageError', 'InvalidArgumentError',
            'NoMemoryError', 'InvalidJobTemplateError',
            'UnsupportedAttributeError', 'InternalError', 'RuntimeError',
            'AlreadyActiveSessionError', 'NoActiveSessionError', 'SessionError',
            'ExitTimeoutError', 'NoResourceUsageError',
            'InvalidAttributeValueError', 'InvalidAttributeFormatError',
            'ConflictingAttributeValuesError', 'InvalidAttributeError',
            'InvalidJobError', 'DrmsInitError', 'SuspendInconsistentStateError',
            'ResumeInconsistentStateError', 'ReleaseInconsistentStateError',
            'HoldInconsistentStateError', 'InconsistentStateError',
            'DrmsExitError', 'DrmCommunicationError', 'DeniedByDrmError',
            'DefaultContactStringError', 'InvalidContactStringError',
            'AuthorizationError', 'ExitError', 'InitError', 'DrmaaError',
            'Session', 'JobTemplate', 'JobInfo' ]
